@@ -2,6 +2,7 @@ import re
 import math
 
 import six
+from six.moves import xrange
 
 from qrcode import base, exceptions
 
@@ -89,6 +90,14 @@ G15_MASK = (1 << 14) | (1 << 12) | (1 << 10) | (1 << 4) | (1 << 1)
 PAD0 = 0xEC
 PAD1 = 0x11
 
+# Precompute bit count limits, indexed by error correction level and code size
+_data_count = lambda block: block.data_count
+BIT_LIMIT_TABLE = [
+    [0] + [8*sum(map(_data_count, base.rs_blocks(version, error_correction)))
+           for version in xrange(1, 41)]
+    for error_correction in xrange(4)
+]
+
 
 def BCH_type_info(data):
         d = data << 10
@@ -139,6 +148,13 @@ def mask_func(pattern):
         return lambda i, j: ((i * j) % 3 + (i + j) % 2) % 2 == 0
     raise TypeError("Bad mask pattern: " + pattern)
 
+def mode_sizes_for_version(version):
+    if version < 10:
+        return MODE_SIZE_SMALL
+    elif version < 27:
+        return MODE_SIZE_MEDIUM
+    else:
+        return MODE_SIZE_LARGE
 
 def length_in_bits(mode, version):
     if mode not in (MODE_NUMBER, MODE_ALPHA_NUM, MODE_8BIT_BYTE,
@@ -149,14 +165,7 @@ def length_in_bits(mode, version):
         raise ValueError("Invalid version (was %s, expected 1 to 40)" %
             version)
 
-    if version < 10:
-        mode_size = MODE_SIZE_SMALL
-    elif version < 27:
-        mode_size = MODE_SIZE_MEDIUM
-    else:
-        mode_size = MODE_SIZE_LARGE
-
-    return mode_size[mode]
+    return mode_sizes_for_version(version)[mode]
 
 
 def lost_point(modules):
@@ -164,63 +173,109 @@ def lost_point(modules):
 
     lost_point = 0
 
-    # LEVEL1
+    lost_point = _lost_point_level1(modules, modules_count)
+    lost_point += _lost_point_level2(modules, modules_count)
+    lost_point += _lost_point_level3(modules, modules_count)
+    lost_point += _lost_point_level4(modules, modules_count)
 
-    for row in range(modules_count):
+    return lost_point
 
-        for col in range(modules_count):
+
+def _lost_point_level1(modules, modules_count):
+    lost_point = 0
+
+    modules_range = xrange(modules_count)
+    row_range_first = (0, 1)
+    row_range_last = (-1, 0)
+    row_range_standard = (-1, 0, 1)
+
+    col_range_first = ((0, 1), (1,))
+    col_range_last = ((-1, 0), (-1,))
+    col_range_standard = ((-1, 0, 1), (-1, 1))
+
+    for row in modules_range:
+
+        if row == 0:
+            row_range = row_range_first
+        elif row == modules_count-1:
+            row_range = row_range_last
+        else:
+            row_range = row_range_standard
+
+        for col in modules_range:
 
             sameCount = 0
             dark = modules[row][col]
 
-            for r in range(-1, 2):
+            if col == 0:
+                col_range = col_range_first
+            elif col == modules_count-1:
+                col_range = col_range_last
+            else:
+                col_range = col_range_standard
 
-                if row + r < 0 or modules_count <= row + r:
-                    continue
+            for r in row_range:
 
-                for c in range(-1, 2):
+                row_offset = row + r
 
-                    if col + c < 0 or modules_count <= col + c:
-                        continue
-                    if r == 0 and c == 0:
-                        continue
+                if r != 0:
+                    col_idx = 0
+                else:
+                    col_idx = 1
 
-                    if dark == modules[row + r][col + c]:
+                for c in col_range[col_idx]:
+
+                    if dark == modules[row_offset][col + c]:
                         sameCount += 1
+
             if sameCount > 5:
                 lost_point += (3 + sameCount - 5)
 
-    # LEVEL2
+    return lost_point
 
-    for row in range(modules_count - 1):
-        for col in range(modules_count - 1):
+
+def _lost_point_level2(modules, modules_count):
+    lost_point = 0
+
+    modules_range = xrange(modules_count - 1)
+
+    for row in modules_range:
+        this_row = modules[row]
+        next_row = modules[row+1]
+        for col in modules_range:
             count = 0
-            if modules[row][col]:
+            if this_row[col]:
                 count += 1
-            if modules[row + 1][col]:
+            if next_row[col]:
                 count += 1
-            if modules[row][col + 1]:
+            if this_row[col + 1]:
                 count += 1
-            if modules[row + 1][col + 1]:
+            if next_row[col + 1]:
                 count += 1
             if count == 0 or count == 4:
                 lost_point += 3
 
-    # LEVEL3
+    return lost_point
 
-    for row in range(modules_count):
-        for col in range(modules_count - 6):
-            if (modules[row][col]
-                    and not modules[row][col + 1]
-                    and modules[row][col + 2]
-                    and modules[row][col + 3]
-                    and modules[row][col + 4]
-                    and not modules[row][col + 5]
-                    and modules[row][col + 6]):
+
+def _lost_point_level3(modules, modules_count):
+    modules_range_short = xrange(modules_count-6)
+
+    lost_point = 0
+    for row in xrange(modules_count):
+        this_row = modules[row]
+        for col in modules_range_short:
+            if (this_row[col]
+                    and not this_row[col + 1]
+                    and this_row[col + 2]
+                    and this_row[col + 3]
+                    and this_row[col + 4]
+                    and not this_row[col + 5]
+                    and this_row[col + 6]):
                 lost_point += 40
 
-    for col in range(modules_count):
-        for row in range(modules_count - 6):
+    for col in xrange(modules_count):
+        for row in modules_range_short:
             if (modules[row][col]
                     and not modules[row + 1][col]
                     and modules[row + 2][col]
@@ -230,21 +285,23 @@ def lost_point(modules):
                     and modules[row + 6][col]):
                 lost_point += 40
 
-    # LEVEL4
-
-    darkCount = 0
-
-    for col in range(modules_count):
-        for row in range(modules_count):
-            if modules[row][col]:
-                darkCount += 1
-
-    ratio = abs(100 * darkCount / modules_count / modules_count - 50) / 5
-    lost_point += ratio * 10
-
     return lost_point
 
 
+def _lost_point_level4(modules, modules_count):
+    modules_range = xrange(modules_count)
+    dark_count = 0
+
+    for row in modules_range:
+        this_row = modules[row]
+        for col in modules_range:
+            if this_row[col]:
+                dark_count += 1
+
+    ratio = abs(100 * dark_count / modules_count / modules_count - 50) / 5
+    return ratio * 10
+
+
 def optimal_data_chunks(data, minimum=4):
     """
     An iterator returning QRData chunks optimized to the data content.
@@ -285,12 +342,11 @@ def _optimal_split(data, pattern):
 
 def to_bytestring(data):
     """
-    Convert data to a (utf-8 encoded) byte-string.
+    Convert data to a (utf-8 encoded) byte-string if it isn't a byte-string
+    already.
     """
-    if not isinstance(data, six.string_types):
-        data = six.text_type(data)
-    if isinstance(data, six.text_type):
-        data = data.encode('utf-8')
+    if not isinstance(data, six.binary_type):
+        data = six.text_type(data).encode('utf-8')
     return data
 
 
@@ -338,12 +394,12 @@ class QRData:
 
     def write(self, buffer):
         if self.mode == MODE_NUMBER:
-            for i in six.moves.xrange(0, len(self.data), 3):
+            for i in xrange(0, len(self.data), 3):
                 chars = self.data[i:i + 3]
                 bit_length = NUMBER_LENGTH[len(chars)]
                 buffer.put(int(chars), bit_length)
         elif self.mode == MODE_ALPHA_NUM:
-            for i in six.moves.xrange(0, len(self.data), 2):
+            for i in xrange(0, len(self.data), 2):
                 chars = self.data[i:i + 2]
                 if len(chars) > 1:
                     buffer.put(ALPHA_NUM.find(chars[0]) * 45 +
@@ -361,7 +417,7 @@ class QRData:
                 buffer.put(c, 8)
 
     def __repr__(self):
-        return self.data
+        return repr(self.data)
 
 
 class BitBuffer:
